# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio
import json
import logging
from concurrent.futures import CancelledError
from pathlib import Path

import websockets

from . import errors, tag, utils
from .client import client, connector
from .errors import JujuAPIError
from .offerendpoints import ParseError as OfferParseError
from .offerendpoints import parse_offer_endpoint, parse_offer_url
from .user import User

log = logging.getLogger(__name__)


class RemoveError(Exception):
    def __init__(self, message):
        self.message = message


class Controller:
    def __init__(
        self,
        max_frame_size=None,
        bakery_client=None,
        jujudata=None,
    ):
        """Instantiate a new Controller.

        One of the connect_* methods will need to be called before this
        object can be used for anything interesting.

        If jujudata is None, jujudata.FileJujuData will be used.

        :param max_frame_size: See
            `juju.client.connection.Connection.MAX_FRAME_SIZE`
        :param bakery_client httpbakery.Client: The bakery client to use
            for macaroon authorization.
        :param jujudata JujuData: The source for current controller
        information.
        """
        self._connector = connector.Connector(
            max_frame_size=max_frame_size,
            bakery_client=bakery_client,
            jujudata=jujudata,
        )
        self._controller_name = None

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def connect(self, *args, **kwargs):
        """Connect to a Juju controller.

        This supports two calling conventions:

        The controller and (optionally) authentication information can be
        taken from the data files created by the Juju CLI.  This convention
        will be used if a ``controller_name`` is specified, or if the
        ``endpoint`` is not.

        Otherwise, both the ``endpoint`` and authentication information
        (``username`` and ``password``, or ``bakery_client`` and/or
        ``macaroons``) are required.

        If a single positional argument is given, it will be assumed to be
        the ``controller_name``.  Otherwise, the first positional argument,
        if any, must be the ``endpoint``.

        Available parameters are:

        :param str controller_name:  Name of controller registered with the
            Juju CLI.
        :param str endpoint: The hostname:port of the controller to connect to.
        :param str username: The username for controller-local users (or None
            to use macaroon-based login.)
        :param str password: The password for controller-local users.
        :param str cacert: The CA certificate of the controller
            (PEM formatted).
        :param httpbakery.Client bakery_client: The macaroon bakery client to
            to use when performing macaroon-based login. Macaroon tokens
            acquired when logging will be saved to bakery_client.cookies.
            If this is None, a default bakery_client will be used.
        :param list macaroons: List of macaroons to load into the
            ``bakery_client``.
        :param int max_frame_size: The maximum websocket frame size to allow.
        :param specified_facades: (deprecated) overwrite the facades with a series of
            specified facades.
        """
        await self.disconnect()
        if "endpoint" not in kwargs and len(args) < 2:
            if args and "model_name" in kwargs:
                raise TypeError("connect() got multiple values for controller_name")
            elif args:
                controller_name = args[0]
            else:
                controller_name = kwargs.pop("controller_name", None)
            await self._connector.connect_controller(controller_name, **kwargs)
        else:
            if "controller_name" in kwargs:
                raise TypeError(
                    "connect() got values for both controller_name and endpoint"
                )
            if args and "endpoint" in kwargs:
                raise TypeError("connect() got multiple values for endpoint")
            has_userpass = len(args) >= 3 or {"username", "password"}.issubset(kwargs)
            has_macaroons = len(args) >= 5 or not {
                "bakery_client",
                "macaroons",
            }.isdisjoint(kwargs)
            if not (has_userpass or has_macaroons):
                raise TypeError("connect() missing auth params")
            arg_names = [
                "endpoint",
                "username",
                "password",
                "cacert",
                "bakery_client",
                "macaroons",
                "max_frame_size",
            ]
            for i, arg in enumerate(args):
                kwargs[arg_names[i]] = arg
            if "endpoint" not in kwargs:
                raise ValueError("endpoint is required if controller_name not given")
            if not (
                {"username", "password"}.issubset(kwargs)
                or {"bakery_client", "macaroons"}.intersection(kwargs)
            ):
                raise ValueError(
                    "Authentication parameters are required "
                    "if controller_name not given"
                )
            await self._connector.connect(**kwargs)
        await self.update_endpoints()

    async def update_endpoints(self):
        try:
            info = await self.info()
            self._connector._connection.endpoints = [
                (e, info.results[0].cacert) for e in info.results[0].addresses
            ]
        except errors.JujuPermissionError:
            log.warning(
                "This user doesn't have at least read access to the controller model, so endpoints are not updated after connection."
            )
            pass

    async def connect_current(self):
        """.. deprecated:: 0.7.3
        Use :meth:`.connect()` instead.
        """
        return await self.connect()

    async def connect_controller(self, controller_name):
        """.. deprecated:: 0.7.3
        Use :meth:`.connect(controller_name)` instead.
        """
        return await self.connect(controller_name)

    async def _connect_direct(self, **kwargs):
        await self.disconnect()
        await self._connector.connect(**kwargs)

    def is_connected(self):
        """Reports whether the Controller is currently connected."""
        return self._connector.is_connected()

    def connection(self):
        """Return the current Connection object. It raises an exception
        if the Controller is disconnected
        """
        return self._connector.connection()

    @property
    def controller_name(self):
        if not self._controller_name:
            try:
                self._controller_name = (
                    self._connector.jujudata.controller_name_by_endpoint(
                        self._connector.connection().endpoint
                    )
                )
            except FileNotFoundError:
                raise errors.PylibjujuError(
                    "Unable to determine controller name. controllers.yaml not found."
                )
        return self._controller_name

    @property
    def controller_uuid(self):
        return self._connector.controller_uuid

    @property
    async def api_endpoints(self):
        """Get API endpoints

        :return list string: List of API Endpoints
        """
        info = await self.info()
        return info.results[0].addresses

    async def disconnect(self):
        """Shut down the watcher task and close websockets."""
        await self._connector.disconnect(entity="controller")

    async def add_credential(
        self, name=None, credential=None, cloud=None, owner=None, force=False
    ):
        """Add or update a credential to the controller.

        :param str name: Name of new credential. If None, the default
            local credential is used.  Name must be provided if a credential
            is given.
        :param CloudCredential credential: Credential to add. If not given,
            it will attempt to read from local data, if available.
        :param str cloud: Name of cloud to associate the credential with.
            Defaults to the same cloud as the controller.
        :param str owner: Username that will own the credential. Defaults to
            the current user.
        :param bool force: Force indicates whether the update should be forced.
            It's only supported for facade v3 or later.
            Defaults to false.
        :returns: Name of credential that was uploaded.
        """
        if not cloud:
            cloud = await self.get_cloud()

        if not owner:
            owner = self.connection().info["user-info"]["identity"]

        if credential and not name:
            raise errors.JujuError("Name must be provided for credential")

        if not credential:
            name, credential = self._connector.jujudata.load_credential(cloud, name)
            if credential is None:
                raise errors.JujuError(f"Unable to find credential: {name}")

        if credential.auth_type == "jsonfile" and "file" in credential.attrs:
            # file creds have to be loaded before being sent to the controller
            try:
                # it might already be JSON
                json.loads(credential.attrs["file"])
            except json.JSONDecodeError:
                # not valid JSON, so maybe it's a file
                cred_path = Path(credential.attrs["file"])
                if cred_path.exists():
                    # make a copy
                    cred_json = credential.to_json()
                    credential = client.CloudCredential.from_json(cred_json)
                    # inline the cred
                    credential.attrs["file"] = cred_path.read_text()

        log.debug("Uploading credential %s", name)
        cloud_facade = client.CloudFacade.from_connection(self.connection())
        tagged_credentials = [
            client.TaggedCredential(
                tag=tag.credential(cloud, tag.untag("user-", owner), name),
                credential=credential,
            )
        ]
        if cloud_facade.version >= 3:
            # UpdateCredentials was renamed to UpdateCredentialsCheckModels
            # in facade version 3.
            await cloud_facade.UpdateCredentialsCheckModels(
                credentials=tagged_credentials,
                force=force,
            )
        else:
            await cloud_facade.UpdateCredentials(credentials=tagged_credentials)
        return name

    async def add_cloud(self, name, cloud):
        """Add a cloud to this controller.

        :param str name: Name to give the new cloud.
        :param Cloud cloud: Cloud configuration.
        :return Cloud: Cloud that was created.
        """
        log.debug("Adding cloud %s", name)
        cloud_facade = client.CloudFacade.from_connection(self.connection())
        await cloud_facade.AddCloud(cloud=cloud, name=name)
        result = await self.cloud(name=name)
        return result.cloud

    async def info(self):
        """Show Controller Info from connection

        :return ControllerAPIInfoResult
        """
        log.debug("Getting information")
        uuids = await self.model_uuids()
        if "controller" not in uuids:
            raise errors.JujuPermissionError("Requires access to controller model.")
        controller_facade = client.ControllerFacade.from_connection(self.connection())
        params = [client.Entity(tag.model(uuids["controller"]))]
        return await controller_facade.ControllerAPIInfoForModels(entities=params)

    async def remove_cloud(self, name):
        """Remove a cloud from this controller.

        :param str name: Name of the cloud to remove.
        """
        log.debug("Removing cloud %s", name)
        cloud_facade = client.CloudFacade.from_connection(self.connection())
        await cloud_facade.RemoveClouds(entities=[client.Entity(tag.cloud(name))])

    async def add_model(
        self,
        model_name,
        cloud_name=None,
        credential_name=None,
        owner=None,
        config=None,
        region=None,
    ):
        """Add a model to this controller.

        :param str model_name: Name to give the new model.
        :param str cloud_name: Name of the cloud in which to create the
            model, e.g. 'aws'. Defaults to same cloud as controller.
        :param str credential_name: Name of the credential to use when
            creating the model. If not given, it will attempt to find a
            default credential.
        :param str owner: Username that will own the model. Defaults to
            the current user.
        :param dict config: Model configuration.
        :param str region: Region in which to create the model.
        :return Model: A connection to the newly created model.
        """
        model_facade = client.ModelManagerFacade.from_connection(self.connection())

        owner = owner or self.connection().info["user-info"]["identity"]
        cloud_name = cloud_name or await self.get_cloud()

        try:
            # attempt to add/update the credential from local data if available
            credential_name = await self.add_credential(
                name=credential_name, cloud=cloud_name, owner=owner
            )
        except errors.JujuError:
            # if it's not available locally, assume it's on the controller
            pass

        if credential_name:
            credential = tag.credential(
                cloud_name, tag.untag("user-", owner), credential_name
            )
        else:
            credential = None

        log.debug("Creating model %s", model_name)

        if not config or "authorized-keys" not in config:
            config = config or {}
            config["authorized-keys"] = await utils.read_ssh_key()

        model_info = await model_facade.CreateModel(
            cloud_tag=tag.cloud(cloud_name),
            config=config,
            credential=credential,
            name=model_name,
            owner_tag=owner,
            region=region,
        )
        from juju.model import Model

        model = Model(jujudata=self._connector.jujudata)
        kwargs = self.connection().connect_params()
        kwargs["uuid"] = model_info.uuid
        model._info = model_info
        await model._connect_direct(**kwargs)

        return model

    async def destroy_models(
        self, *models, destroy_storage=False, force=False, max_wait=None
    ):
        """Destroy one or more models.

        :param str *models: Names or UUIDs of models to destroy
        :param bool destroy_storage: Whether or not to destroy storage when
            destroying the models. Defaults to false.
        :param bool force: Whether or not to force when destroying the models.
            Defaults to false.
        :param int max_wait : Max time in seconds to wait when destroying the models.

        """
        uuids = await self.model_uuids()
        models = [uuids.get(model) or model for model in models]

        model_facade = client.ModelManagerFacade.from_connection(self.connection())

        log.debug(
            "Destroying model%s %s", "" if len(models) == 1 else "s", ", ".join(models)
        )

        if model_facade.version >= 5:
            params = [
                client.DestroyModelParams(
                    model_tag=tag.model(model),
                    destroy_storage=destroy_storage,
                    force=force,
                    max_wait=max_wait,
                )
                for model in models
            ]
            await model_facade.DestroyModels(models=params)
        else:
            params = [client.Entity(tag.model(model)) for model in models]

            await model_facade.DestroyModels(entities=params)

    destroy_model = destroy_models

    async def add_user(self, username, password=None, display_name=None):
        """Add a user to this controller.

        :param str username: Username
        :param str password: Password
        :param str display_name: Display name
        :returns: A :class:`~juju.user.User` instance
        """
        if not display_name:
            display_name = username
        user_facade = client.UserManagerFacade.from_connection(self.connection())
        users = [
            client.AddUser(
                display_name=display_name, username=username, password=password
            )
        ]
        results = await user_facade.AddUser(users=users)
        secret_key = results.results[0].secret_key
        return await self.get_user(username, secret_key=secret_key)

    async def remove_user(self, username):
        """Remove a user from this controller."""
        client_facade = client.UserManagerFacade.from_connection(self.connection())
        user = tag.user(username)
        await client_facade.RemoveUser(entities=[client.Entity(user)])

    async def change_user_password(self, username, password):
        """Change the password for a user in this controller.

        :param str username: Username
        :param str password: New password

        """
        user_facade = client.UserManagerFacade.from_connection(self.connection())
        entity = client.EntityPassword(password=password, tag=tag.user(username))
        return await user_facade.SetPassword(changes=[entity])

    async def reset_user_password(self, username):
        """Reset user password.

        :param str username: Username
        :returns: A :class:`~juju.user.User` instance
        """
        user_facade = client.UserManagerFacade.from_connection(self.connection())
        entity = client.Entity(tag.user(username))
        results = await user_facade.ResetPassword(entities=[entity])
        secret_key = results.results[0].secret_key
        return await self.get_user(username, secret_key=secret_key)

    async def destroy(self, destroy_all_models=False, destroy_storage=False):
        """Destroy this controller.

        :param bool destroy_all_models: Destroy all hosted models in the
            controller.
        :param bool destroy_storage: Destroy all hosted storage in the
            controller.
        """
        controller_facade = client.ControllerFacade.from_connection(self.connection())
        return await controller_facade.DestroyController(
            destroy_models=destroy_all_models, destroy_storage=destroy_storage
        )

    async def disable_user(self, username):
        """Disable a user.

        :param str username: Username

        """
        user_facade = client.UserManagerFacade.from_connection(self.connection())
        entity = client.Entity(tag.user(username))
        return await user_facade.DisableUser(entities=[entity])

    async def enable_user(self, username):
        """Re-enable a previously disabled user."""
        user_facade = client.UserManagerFacade.from_connection(self.connection())
        entity = client.Entity(tag.user(username))
        return await user_facade.EnableUser(entities=[entity])

    async def get_model_info(self, model_name=None, model_uuid=None):
        """Return a client.ModelInfo object for a given Model.

        Retrieves latest info for this Model from the api server. The
        return value is cached on the Model.info attribute so that the
        valued may be accessed again without another api call, if
        desired.

        This method is called automatically when the Model is connected,
        resulting in Model.info being initialized without requiring an
        explicit call to this method.

        """
        if model_uuid is None and model_name is None:
            raise errors.JujuError(
                "get_model_info requires either a name or a uuid for a model"
            )

        facade = client.ModelManagerFacade.from_connection(self.connection())
        if model_uuid is None:
            uuids = await self.model_uuids()
            try:
                model_uuid = uuids[model_name]
            except KeyError:
                raise errors.JujuError(
                    f"{model_name} is not among the models in the controller : {uuids}"
                )
        entity = client.Entity(tag.model(model_uuid))
        _model_info_results = await facade.ModelInfo(entities=[entity])
        return _model_info_results.results[0].result

    async def cloud(self, name=None):
        """Get Cloud

        :param str name: Cloud name. If not specified, the cloud where
                         the controller lives on is returned.
        :returns: -> ~CloudResult
        """
        if name is None:
            name = await self.get_cloud()
        entity = client.Entity(tag.cloud(name))
        cloud_facade = client.CloudFacade.from_connection(self.connection())
        cloud = await cloud_facade.Cloud(entities=[entity])
        if len(cloud.results) == 0:
            log.error("No clouds found.")
            raise
        elif len(cloud.results) > 1:
            log.error("More than one cloud found.")
            raise
        return cloud.results[0]

    async def clouds(self):
        """Get all the clouds in the controller

        :returns: -> ~CloudsResult
        """
        cloud_facade = client.CloudFacade.from_connection(self.connection())
        return await cloud_facade.Clouds()

    async def get_cloud(self):
        """Get the name of the cloud that this controller lives on."""
        cloud_facade = client.CloudFacade.from_connection(self.connection())

        result = await cloud_facade.Clouds()
        cloud = next(iter(result.clouds))
        return tag.untag("cloud-", cloud)

    async def get_models(self, all=False, username=None):  # noqa: A002
        """.. deprecated:: 0.7.0
        Use :meth:`.list_models` instead.
        """
        return await self.list_models(username, all)

    async def model_uuids(self, username=None, all=False):  # noqa: A002
        """Return a mapping of model names to UUIDs the given user can access.

        :param str username: Optional username argument, defaults to
        current connected user.

        :param bool all: Flag to list all models, regardless of
        user accessibility (administrative users only)

        :returns: {str name : str UUID}
        """
        model_manager_facade = client.ModelManagerFacade.from_connection(
            self.connection()
        )
        u_name = username if username else self.get_current_username()
        user = tag.user(u_name)

        user_model_list = await model_manager_facade.ListModelSummaries(
            user_tag=user, all_=all
        )
        model_summaries = [msr.result for msr in user_model_list.results]
        return {
            model_summary.name: model_summary.uuid for model_summary in model_summaries
        }

    async def list_models(self, username=None, all=False):  # noqa: A002
        """Return list of names of the available models on this controller.

        Equivalent to ``sorted((await self.model_uuids()).keys())``
        """
        uuids = await self.model_uuids(username, all)
        return sorted(uuids.keys())

    async def get_current_user(self, secret_key=None):
        """Returns the user object associated with the current connection.
        :param str secret_key: Issued by juju when add or reset user
            password

        :returns: A :class:`~juju.user.User` instance
        """
        return await self.get_user(self.connection().username)

    def get_current_username(self):
        """Returns the username associated with the current connection.

        :returns: :str: username of the connected user
        """
        return self.connection().username

    async def get_model(self, model):
        """Get a model by name or UUID.

        :param str model: Model name or UUID
        :returns Model: Connected Model instance.
        """
        uuids = await self.model_uuids()
        uuid = uuids.get(model) or model

        from juju.model import Model

        model = Model()
        kwargs = self.connection().connect_params()
        kwargs["uuid"] = uuid
        await model._connect_direct(**kwargs)
        return model

    async def get_user(self, username, secret_key=None):
        """Get a user by name.

        :param str username: Username
        :param str secret_key: Issued by juju when add or reset user
            password
        :returns: A :class:`~juju.user.User` instance
        """
        client_facade = client.UserManagerFacade.from_connection(self.connection())
        user = tag.user(username)
        args = [client.Entity(user)]
        try:
            response = await client_facade.UserInfo(
                entities=args, include_disabled=True
            )
        except errors.JujuError as e:
            if "permission denied" in e.errors:
                # apparently, trying to get info for a nonexistent user returns
                # a "permission denied" error rather than an empty result set
                return None
            raise
        if response.results and response.results[0].result:
            return User(self, response.results[0].result, secret_key=secret_key)
        return None

    async def get_users(self, include_disabled=False):
        """Return list of users that can connect to this controller.

        :param bool include_disabled: Include disabled users
        :returns: A list of :class:`~juju.user.User` instances
        """
        client_facade = client.UserManagerFacade.from_connection(self.connection())
        response = await client_facade.UserInfo(
            entities=None, include_disabled=include_disabled
        )
        return [User(self, r.result) for r in response.results]

    async def grant(self, username, acl="login"):
        """Grant access level of the given user on the controller.
        Note that if the user already has higher permissions than the
        provided ACL, this will do nothing (see revoke for a way to
        remove permissions).
        :param str username: Username
        :param str acl: Access control ('login', 'add-model' or 'superuser')
        :returns: True if new access was granted, False if user already had
            requested access or greater.  Raises JujuError if failed.
        """
        controller_facade = client.ControllerFacade.from_connection(self.connection())
        user = tag.user(username)
        changes = client.ModifyControllerAccess(acl, "grant", user)
        try:
            await controller_facade.ModifyControllerAccess(changes=[changes])
            return True
        except errors.JujuError as e:
            if "user already has" in str(e):
                return False
            else:
                raise

    async def revoke(self, username, acl="login"):
        """Removes some or all access of a user to from a controller
        If 'login' access is revoked, the user will no longer have any
        permissions on the controller. Revoking a higher privilege from
        a user without that privilege will have no effect.

        :param str username: username
        :param str acl: Access to remove ('login', 'add-model' or 'superuser')
        """
        controller_facade = client.ControllerFacade.from_connection(self.connection())
        user = tag.user(username)
        changes = client.ModifyControllerAccess(acl, "revoke", user)
        return await controller_facade.ModifyControllerAccess(changes=[changes])

    async def grant_model(self, username, model_uuid, acl="read"):
        """Grant a user access to a model. Note that if the user
        already has higher permissions than the provided ACL,
        this will do nothing (see revoke_model for a way to remove
        permissions).

        :param str username: Username
        :param str model_uuid: The UUID of the model to change.
        :param str acl: Access control ('read, 'write' or 'admin')
        """
        model_facade = client.ModelManagerFacade.from_connection(self.connection())
        user = tag.user(username)
        model = tag.model(model_uuid)
        changes = client.ModifyModelAccess(acl, "grant", model, user)
        return await model_facade.ModifyModelAccess(changes=[changes])

    async def revoke_model(self, username, model_uuid, acl="read"):
        """Revoke some or all of a user's access to a model.
        If 'read' access is revoked, the user will no longer have any
        permissions on the model. Revoking a higher privilege from
        a user without that privilege will have no effect.

        :param str username: Username to revoke
        :param str model_uuid: The UUID of the model to change.
        :param str acl: Access control ('read, 'write' or 'admin')
        """
        model_facade = client.ModelManagerFacade.from_connection(self.connection())
        user = tag.user(username)
        model = tag.model(model_uuid)
        changes = client.ModifyModelAccess(acl, "revoke", model, user)
        return await model_facade.ModifyModelAccess(changes=[changes])

    async def create_offer(
        self, model_uuid, endpoint, offer_name=None, application_name=None
    ):
        """Offer a deployed application using a series of endpoints for use by
        consumers.

        @param endpoint: holds the application and endpoint you want to offer
        @param offer_name: override the offer name to help the consumer
        @param application_name: overrides the application name in the endpoint
        """
        # If we have both the offer_name and the application_name
        # then we're coming from bundle/overlays, so no need to parse the endpoint
        # Also we accept endpoints without a colon (:) in the overlays
        if offer_name and application_name:
            o_name = offer_name
            a_name = application_name
            eps = {endpoint: endpoint}
        else:
            try:
                offer = parse_offer_endpoint(endpoint)
            except OfferParseError as e:
                log.error(e.message)
                raise

            o_name = offer_name if offer_name else offer.application
            a_name = application_name if application_name else offer.application
            eps = {name: name for name in offer.endpoints}

        params = client.AddApplicationOffer()
        params.application_name = a_name
        params.endpoints = eps
        params.offer_name = o_name
        params.model_tag = tag.model(model_uuid)

        facade = client.ApplicationOffersFacade.from_connection(self.connection())
        return await facade.Offer(offers=[params])

    async def list_offers(self, model_name):
        """Offers list information about applications' endpoints that have been
        shared and who is connected.
        """
        params = client.OfferFilter()
        params.model_name = model_name

        facade = client.ApplicationOffersFacade.from_connection(self.connection())
        return await facade.ListApplicationOffers(filters=[params])

    async def remove_offer(self, model_uuid, offer, force=False):
        """Remove offer for an application.

        Offers will also remove relations to those offers, use force to do
        so, without an error.
        """
        url = None
        try:
            url = parse_offer_url(offer)
        except OfferParseError as e:
            log.error(e.message)
            raise
        if url is None:
            raise Exception

        offer_source = url.source
        if offer_source == "":
            offer_source = self.controller_name
        if not force:
            raise RemoveError(
                "removing offer will also remove relations, use force and try again."
            )

        facade = client.ApplicationOffersFacade.from_connection(self.connection())
        return await facade.DestroyOffers(force=force, offer_urls=[url.string()])

    async def get_consume_details(self, endpoint):
        """get_consume_details returns the details necessary to pass to another
        model to consume the specified offers represented by the urls.
        """
        facade = client.ApplicationOffersFacade.from_connection(self.connection())
        offers = await facade.GetConsumeDetails(
            offer_urls=client.OfferURLs(offer_urls=[endpoint])
        )
        if len(offers.results) != 1:
            raise JujuAPIError("expected to find one result")
        result = offers.results[0]
        if result.error is not None:
            raise JujuAPIError(result.error)

        return result

    async def watch_model_summaries(self, callback, as_admin=False):
        """Watch the controller for model summary updates.

        If as_admin is true, a call will be made as the admin to watch
        all models in the controller. If the user isn't a superuser they
        will get a permission error.
        """
        stop_event = asyncio.Event()

        async def _watcher(stop_event):
            try:
                facade = client.ControllerFacade.from_connection(self.connection())
                watcher = client.ModelSummaryWatcherFacade.from_connection(
                    self.connection()
                )
                if as_admin:
                    result = await facade.WatchAllModelSummaries()
                    watcher.Id = result.watcher_id
                else:
                    result = await facade.WatchModelSummaries()
                    log.debug(f"watcher id: {result.watcher_id}")
                    watcher.Id = result.watcher_id

                while True:
                    try:
                        results = await utils.run_with_interrupt(
                            watcher.Next(), stop_event, log=log
                        )
                    except JujuAPIError as e:
                        if "watcher was stopped" not in str(e):
                            raise
                    except websockets.ConnectionClosed:
                        break
                    if stop_event.is_set():
                        try:
                            await watcher.Stop()
                        except websockets.ConnectionClosed:
                            pass  # can't stop on a closed conn
                        break
                    for summary in results.models:
                        callback(summary)
            except CancelledError:
                pass
            except Exception:
                log.exception("Error in watcher")
                raise

        log.debug("Starting watcher task for model summaries")
        asyncio.ensure_future(_watcher(stop_event))  # noqa: RUF006
        return stop_event

    async def add_secret_backends(self, id_, name, backend_type, config):
        """Add a new secret backend.

        Parameters
        ----------
        id : string
            id for the backend
        name : string
            name of the backend
        backend-type : string
        config : dict
            dictionary with the backend configuration values

        Returns
        -------
        list
           a list of errors if any

        """
        facade = client.SecretBackendsFacade.from_connection(self.connection())
        return await facade.AddSecretBackends([
            {
                "id": id_,
                "backend-type": backend_type,
                "config": config,
                "name": name,
                "token-rotate-interval": config.get("token-rotate-interval", None),
            }
        ])

    async def list_secret_backends(self, reveal=False):
        """Return the list of secret backends

        Parameters
        ----------
        reveal : boolean
            include sensitive backend config content if true

        Returns
        -------
        list
           a list of available secret backends

        """
        facade = client.SecretBackendsFacade.from_connection(self.connection())
        return await facade.ListSecretBackends(None, reveal)

    async def remove_secret_backends(self, name, force=False):
        """Remove a secrets backend.

        Parameters
        ----------
        name : name of the backend
        force : true if the operation is foced

        Returns
        -------
        error if any

        """
        facade = client.SecretBackendsFacade.from_connection(self.connection())
        return await facade.RemoveSecretBackends([{"name": name, "force": force}])

    async def update_secret_backends(
        self,
        name,
        config=None,
        force=False,
        name_change=None,
        token_rotate_interval=None,
    ):
        """Update a backend.

        Parameters
        ----------
        name : string
            the backend name
        config : dict
            key value dict with configuration parameters
        force : boolean
            true to force the update process
        name_change : string
            new name for the backend
        token_rotate_interval : int
            token rotation interval

        """
        facade = client.SecretBackendsFacade.from_connection(self.connection())
        return await facade.UpdateSecretBackends([
            {
                "name": name,
                "config": config,
                "force": force,
                "token-rotate-interval": token_rotate_interval,
                "name-change": name_change,
            }
        ])


class ConnectedController(Controller):
    def __init__(
        self,
        connection,
        max_frame_size=None,
        bakery_client=None,
        jujudata=None,
    ):
        super().__init__(
            max_frame_size=max_frame_size,
            bakery_client=bakery_client,
            jujudata=jujudata,
        )
        self._conn = connection

    async def __aenter__(self):
        kwargs = self._conn.connect_params()
        kwargs.pop("uuid")
        await self._connect_direct(**kwargs)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.


class JujuError(Exception):
    def __init__(self, *args, **kwargs):
        self.message = ""
        self.errors = []
        if args:
            self.message = str(args[0])
            if isinstance(args[0], (list, tuple)):
                self.errors = args[0]
            elif len(args) > 1:
                self.errors = list(args)
            else:
                self.errors = [self.message]
        super().__init__(*args, **kwargs)


class JujuAPIError(JujuError):
    def __init__(self, result):
        self.result = result
        self.message = result["error"]
        self.error_code = result.get("error-code")
        self.response = result["response"]
        self.request_id = result["request-id"]
        self.error_info = result.get("error-info")
        super().__init__(self.message)


class JujuConnectionError(ConnectionError, JujuError):
    pass


class JujuAuthError(JujuConnectionError):
    pass


class JujuRedirectException(Exception):
    """Exception indicating that a redirection was requested"""

    def __init__(self, redirect_info, follow_redirect=True):
        self.redirect_info = redirect_info
        self.follow_redirect = follow_redirect

    @property
    def ca_cert(self):
        return self.redirect_info["ca-cert"]

    @property
    def endpoints(self):
        return [
            ("{value}:{port}".format(**s), self.ca_cert)
            for servers in self.redirect_info["servers"]
            for s in servers
            if s["scope"] == "public" or not self.follow_redirect
        ]


class JujuEntityNotFoundError(JujuError):
    """Exception indicating that an entity was not found in the state. It was
    expected that the entity was found in state and this is a terminal
    condition.
    To fix this condition, you should disconnect and reconnect to ensure that
    any missing entities are correctly picked up.
    """

    def __init__(self, entity_name, entity_types=None):
        self.entity_name = entity_name
        self.entity_types = entity_types
        super().__init__(f"Entity not found: {entity_name}")


class JujuModelError(JujuError):
    pass


class JujuMachineError(JujuError):
    pass


class JujuAgentError(JujuError):
    pass


class JujuAppError(JujuError):
    pass


class JujuUnitError(JujuError):
    pass


class JujuPermissionError(JujuError):
    pass


class JujuBackupError(JujuError):
    pass


class PylibjujuProgrammingError(Exception):
    pass


class JujuNotSupportedError(JujuError):
    pass


class JujuNotValid(JujuError):
    def __init__(self, entity_type, entity_name):
        self.entity_type = entity_type
        self.entity_name = entity_name
        super().__init__(f"Invalid {entity_type} : {entity_name}")


class JujuConfigError(JujuError):
    """Exception raised during processing a configuration key-value pair
    in a config set for an application.
    """

    def __init__(self, config, config_pair, message=None):
        self.config = config
        self.config_pair = config_pair
        if message is None:
            self.message = (
                "Couldn't process the value of a config pair : %s, value of type %s"
                % (self.config_pair, type(self.config_pair[1]))
            )
        else:
            self.message = message
        super().__init__(self.message)


class JujuApplicationConfigError(JujuConfigError):
    pass


class JujuModelConfigError(JujuConfigError):
    pass


class JujuControllerNotFoundError(JujuError):
    pass


class AbstractMethodError(Exception):
    pass


class PylibjujuError(JujuError):
    pass


class JujuUnknownVersion(PylibjujuError):
    pass



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.
from __future__ import annotations

import logging
import sys
import warnings

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

from .client import client

log = logging.getLogger(__name__)


class StatusStr(StrEnum):
    """Recognised status values.

    Please keep this set exact same as the severity map below.
    """

    error = "error"
    blocked = "blocked"
    waiting = "waiting"
    maintenance = "maintenance"
    active = "active"
    terminated = "terminated"
    unknown = "unknown"


""" severity_map holds status values with a severity measure.
Status values with higher severity are used in preference to others.
"""
severity_map: dict[StatusStr, int] = {
    # FIXME: Juju defines a lot more status values #1204
    StatusStr.error: 100,
    StatusStr.blocked: 90,
    StatusStr.waiting: 80,
    StatusStr.maintenance: 70,
    StatusStr.active: 60,
    StatusStr.terminated: 50,
    StatusStr.unknown: 40,
}


def derive_status(statuses: list[str | StatusStr]) -> StatusStr:
    """Derive status from a set.

    derive_status is used to determine the application status from a set of unit
    status values.

    :param statuses: list of known unit workload statuses
    """
    current: StatusStr = StatusStr.unknown
    for status in statuses:
        try:
            status = StatusStr(status)
        except ValueError:
            # Unknown Juju status, let's assume it's least important
            continue

        if severity_map[status] > severity_map[current]:
            current = status
    return current


async def formatted_status(model, target=None, raw=False, filters=None):
    """Returns a string that mimics the content of the information
    returned in the juju status command. If the raw parameter is
    enabled, the function returns a FullStatus object.
    :param Model model: model object to be used
    :param Fileobject target: if set expects a file object such as
        sys.stdout or a file descriptor. The obtained status will
        be sent to the file using the write function. If set to
        `None`, this function returns a string with the formatted
        status.
    :param bool raw: if `true` this functions returns the raw
        `FullStatus` object returned by Juju. This is similar to
        invoking `get_status`.
    :param str filters: Optional list of applications, units, or machines
        to include, which can use wildcards ('*').
    """
    warnings.warn(
        "juju.status.formatted_status is deprecated, the implementation is likely broken",
        DeprecationWarning,
        stacklevel=2,
    )
    client_facade = client.ClientFacade.from_connection(model.connection())
    result_status = await client_facade.FullStatus(patterns=filters)

    if raw:
        result_str = str(result_status)
    else:
        result_str = _print_status_model(result_status)
        result_str += "\n"
        result_str += _print_status_apps(result_status)
        result_str += "\n"
        result_str += _print_status_units(result_status)
        result_str += "\n"
        result_str += _print_status_machines(result_status)
        result_str += "\n"
    if target is None:
        return result_str

    try:
        target.write(result_str)
    except Exception as e:
        logging.error(e)

    return None


def _print_status_model(result_status):
    """Private function to print the status of a model"""
    m = result_status.model
    # print model
    result_str = "{:<25} {:<25} {:<15} {:<15} {:<30} {:<30}\n".format(
        "Model", "Cloud/Region", "Version", "SLA", "Timestamp", "Notes"
    )
    sla = m.unknown_fields["sla"]
    cloud = m.cloud_tag.split("-")[1]
    timestamp = result_status.controller_timestamp
    if m.available_version is not None and m.available_version != "":
        available_version = f"upgrade available: {m.available_version}"
    else:
        available_version = ""
    result_str += "{:<25} {:<25} {:<15} {:<15} {:<30} {:<30}".format(
        m.name, cloud + "/" + m.region, m.version, sla, timestamp, available_version
    )
    result_str += "\n"
    return result_str


def _print_status_apps(result_status):
    """Auxiliary function to print the apps received
    in a status result
    """
    apps = result_status.applications
    if apps is None or len(apps) == 0:
        return ""

    limits = "{:<25} {:<10} {:<10} {:<5} {:<20} {:<8}"
    # print header
    result_str = limits.format("App", "Version", "Status", "Scale", "Charm", "Channel")

    for name, app in apps.items():
        # extract charm name from the path
        # like in ch:amd64/trusty/mediawiki-28
        charm_name = app.charm.split("/")[-1]
        charm_name = charm_name.split("-")[0]
        work_ver = "NA" if app.workload_version is None else app.workload_version
        charm_channel = "NA" if app.charm_channel is None else app.charm_channel
        app_units = "NA" if app.units is None else len(app.units)
        app_status = "NA" if app.status.status is None else app.status.status
        result_str += "\n"
        result_str += limits.format(
            name, work_ver, app_status, app_units, charm_name, charm_channel
        )
    result_str += "\n"
    return result_str


def _print_status_units(result_status):
    """Auxiliary function to print the units received
    in a status result
    """
    apps = result_status.applications
    if apps is None or len(apps) == 0:
        return

    limits = "{:<15} {:<15} {:<20} {:<10} {:<15} {:<10} {:<30}"
    summary = ""
    for app in apps.values():
        units = app.units
        if units is None or len(units) == 0:
            continue

        for name, unit in units.items():
            addr = unit.public_address
            if addr is None:
                addr = ""

            if unit.opened_ports is None:
                opened_ports = ""
            else:
                opened_ports = ",".join(unit.opened_ports)

            info = unit.workload_status.info
            if info is None:
                info = ""

            step = limits.format(
                name,
                unit.workload_status.status,
                unit.agent_status.status,
                unit.machine,
                addr,
                opened_ports,
                info,
            )
            if summary == "":
                summary = step
            else:
                summary = summary + "\n" + step

    if len(summary) == 0:
        return ""
    result_str = limits.format(
        "Unit", "Workload", "Agent", "Machine", "Public address", "Ports", "Message"
    )
    result_str += "\n"
    result_str += summary
    result_str += "\n"
    return result_str


def _print_status_machines(result_status):
    machines = result_status.machines
    if machines is None or len(machines) == 0:
        return

    limits = "{:<15} {:<15} {:<15} {:<20} {:<15} {:<30}"
    summary = ""
    for name, machine in machines.items():
        dns = machine.dns_name
        if dns is None:
            dns = ""
        step = limits.format(
            name,
            machine.agent_status.status,
            dns,
            machine.instance_id,
            machine.series,
            machine.agent_status.info,
        )
        if summary == "":
            summary = step
        else:
            summary = summary + "\n" + step

    if summary == "":
        return
    result_str = limits.format(
        "Machine", "State", "DNS", "Inst id", "Series", "Message"
    )
    result_str += "\n"
    result_str += summary
    result_str += "\n"
    return result_str



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Deploy a charm and waits until it reports itself active
3. Destroys the unit and application

"""

import asyncio

from juju.model import Model


async def main():
    model = Model()
    print("Connecting to model")
    await model.connect()

    try:
        print("Deploying ")
        application = await model.deploy(
            "ch:juju-qa-test",
            application_name="juju-qa-test2",
            channel="2.0/edge",
        )

        print("Waiting for active")
        await model.wait_for_idle(status="active")

        # when run on a container based model it should auto-switch to
        # scale instead of failing with JujuError
        await application.add_unit(count=2)
        await application.destroy()

    finally:
        print("Disconnecting from model")
        await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

from . import model


class Action(model.ModelEntity):
    def __init__(self, entity_id, model, history_index=-1, connected=True):
        super().__init__(entity_id, model, history_index, connected)
        self.results = {}
        self._status = self.data["status"]

    @property
    def status(self):
        return self._status

    async def fetch_output(self):
        completed_action = await self.model._get_completed_action(self.id)
        self.results = completed_action.output or {}
        self._status = completed_action.status

    async def wait(self):
        self.results or await self.fetch_output()
        return self



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import re
from enum import Enum, unique


@unique
class MatchType(Enum):
    MATCH = 1
    SEARCH = 2


MODEL = re.compile("^[a-z0-9]+[a-z0-9-]*$")


def match_model(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(MODEL, val)
    else:
        return re.match(MODEL, val)


APPLICATION = re.compile("^(?:[a-z][a-z0-9]*(?:-[a-z0-9]*[a-z][a-z0-9]*)*)$")


def match_application(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(APPLICATION, val)
    else:
        return re.match(APPLICATION, val)


def is_valid_application(val):
    return match_application(val) is not None


ENDPOINT = re.compile(
    "/?((?P<model>[^\\.]*)\\.)?(?P<appname>[^:]*)(:(?P<endpoints>.*))?"
)


def match_endpoint(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(ENDPOINT, val)
    else:
        return re.match(ENDPOINT, val)


SOURCE_ENDPOINT = re.compile("^[a-zA-Z0-9]+$")


def match_source_endpoint(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(SOURCE_ENDPOINT, val)
    else:
        return re.match(SOURCE_ENDPOINT, val)


MODEL_APPLICATION = re.compile(
    "(/?((?P<user>[^/]+)/)?(?P<model>[^.]*)(.(?P<application>[^:]*(:.*)?))?)?"
)


def match_model_application(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(MODEL_APPLICATION, val)
    else:
        return re.match(MODEL_APPLICATION, val)


valid_user_name_snippet = "[a-zA-Z0-9][a-zA-Z0-9.+-]*[a-zA-Z0-9]"
valid_user_snippet = f"(?:{valid_user_name_snippet}(?:@{valid_user_name_snippet})?)"
USER = re.compile(
    f"^(?P<name>{valid_user_name_snippet})(?:@(?P<domain>{valid_user_name_snippet}))?$"
)


def match_user(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(USER, val)
    else:
        return re.match(USER, val)


RELATION = re.compile("[a-z][a-z0-9]*(?:[_-][a-z0-9]+)*")


def match_relation(val, match_type=None):
    if match_type is MatchType.SEARCH:
        return re.search(RELATION, val)
    else:
        return re.match(RELATION, val)



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Deploys a charm and waits until it reports itself active
3. Creates an offer
4. Lists the offer
3. Destroys the unit and application

"""

import asyncio
import tempfile
from logging import getLogger

from juju.controller import Controller

log = getLogger(__name__)


async def main():
    controller = Controller()
    print("Connecting to controller")
    await controller.connect()

    try:
        print("Creating models")
        offering_model = await controller.add_model("test-cmr-1")
        consuming_model = await controller.add_model("test-cmr-2")

        print("Deploying mysql")
        application = await offering_model.deploy(
            "ch:mysql",
            application_name="mysql",
            series="jammy",
            channel="edge",
        )

        print("Waiting for active")
        await offering_model.block_until(
            lambda: all(unit.workload_status == "active" for unit in application.units)
        )

        print("Adding offer")
        await offering_model.create_offer("mysql:db")

        offers = await offering_model.list_offers()
        print(
            "Show offers",
            ", ".join(
                "%s: %s" % item
                for offer in offers.results
                for item in vars(offer).items()
            ),
        )

        print("Consuming offer")
        await consuming_model.consume("admin/test-cmr-1.mysql")

        print("Exporting bundle")
        with tempfile.TemporaryDirectory() as dirpath:
            await offering_model.export_bundle(f"{dirpath}/bundle.yaml")

        print("Remove SAAS")
        await consuming_model.remove_saas("mysql")

        print("Removing offer")
        await offering_model.remove_offer("admin/test-cmr-1.mysql", force=True)

        print("Destroying models")
        await controller.destroy_model(offering_model.info.uuid)
        await controller.destroy_model(consuming_model.info.uuid)

    except Exception:
        log.exception("Example failed!")
        raise

    finally:
        print("Disconnecting from controller")
        await controller.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example doesn't work - it demonstrates features that don't exist yet."""

import asyncio
import logging

from juju.model import Model


async def main():
    model = Model()
    # connect to current model with current user, per Juju CLI
    await model.connect()

    goal_state = Model.from_yaml("bundle-like-thing")
    ubuntu_app = await model.deploy(
        "ch:ubuntu",
        application_name="ubuntu",
        series="trusty",
        channel="stable",
    )
    ubuntu_app.on_unit_added(callback=lambda unit: True)

    await model.deploy(
        "ch:nrpe",
        application_name="nrpe",
        series="trusty",
        channel="stable",
        num_units=0,
    )
    await model.relate(
        "ubuntu",
        "nrpe",
    )

    _result, _ok = await model.block_until(
        lambda: model.matches(goal_state), timeout=600
    )


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio
import sys

from juju.controller import Controller


async def main(cloud_name, credential_name):
    controller = Controller()
    model = None
    print("Connecting to controller")
    # connect to current controller with current user, per Juju CLI
    await controller.connect()
    try:
        print("Adding model")
        model = await controller.add_model(
            "test", cloud_name=cloud_name, credential_name=credential_name
        )

        # verify credential
        print(f"Verify model's credential: {model.info.cloud_credential_tag}")

        # verify we can deploy
        print("Deploying ubuntu")
        app = await model.deploy("ch:ubuntu")

        print("Waiting for active")
        await model.block_until(
            lambda: app.units
            and all(unit.workload_status == "active" for unit in app.units)
        )

        print("Removing ubuntu")
        await app.remove()
    finally:
        print("Cleaning up")
        if model:
            print("Removing model")
            model_uuid = model.info.uuid
            await model.disconnect()
            await controller.destroy_model(model_uuid)
        print("Disconnecting")
        await controller.disconnect()


if __name__ == "__main__":
    assert len(sys.argv) > 2, "Please provide a cloud and credential name"
    asyncio.run(main(sys.argv[1], sys.argv[2]))



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Deploy a bundle and waits until it reports itself active
3. Destroys the units and applications

"""

import asyncio

from juju.controller import Controller


async def main():
    controller = Controller()
    # connect to current controller with current user, per Juju CLI
    await controller.connect()

    # Deploy charmhub bundle
    await deploy_bundle(controller, "juju-qa-bundle-test")

    await controller.disconnect()


async def deploy_bundle(controller, url, channel=None):
    models = await controller.list_models()
    model = await controller.add_model(f"model{len(models) + 1}")

    try:
        print("Deploying bundle")

        applications = await deploy_and_wait_for_bundle(model, url, channel)

        print("Successfully deployed!")
        print("Removing bundle")
        for application in applications:
            await application.remove()
    finally:
        print("Disconnecting from model")
        await model.disconnect()
        print("Success")


async def deploy_and_wait_for_bundle(model, url, channel=None):
    applications = await model.deploy(url, channel=channel)

    print("Waiting for active")
    await model.block_until(
        lambda: all(
            unit.workload_status == "active"
            for application in applications
            for unit in application.units
        )
    )

    return applications


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

import asyncio

from juju.model import Model


async def status():
    model = Model()
    await model.connect()

    status = await model.get_status()
    await model.disconnect()

    print("Applications:", list(status.applications.keys()))
    print("Machines:", list(status.machines.keys()))
    print("Relations:", status.relations)

    return status


if __name__ == "__main__":
    asyncio.run(status())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Watches the model and prints all changes
3. Runs forever (kill with Ctrl-C)

"""

import asyncio

from juju.model import Model


async def on_model_change(delta, old, new, model):
    print(delta.entity, delta.type, delta.data)
    print(old)
    print(new)
    print(model)
    await asyncio.sleep(0)


async def watch_model():
    model = Model()
    # connect to current model with current user, per Juju CLI
    await model.connect()

    model.add_observer(on_model_change)


if __name__ == "__main__":
    # Run loop until the process is manually stopped (watch_model will loop
    # forever).
    asyncio.run(watch_model())



# DO NOT CHANGE THIS FILE! This file is auto-generated by facade.py.
# Changes will be overwritten/lost when the file is regenerated.

from juju.client._definitions import *
from juju.client.facade import ReturnMapping, Type


class ApplicationFacade(Type):
    name = "Application"
    version = 20

    @ReturnMapping(AddRelationResults)
    async def AddRelation(self, endpoints=None, via_cidrs=None):
        """AddRelation adds a relation between the specified endpoints and returns the relation info.

        endpoints : typing.Sequence[str]
        via_cidrs : typing.Sequence[str]
        Returns -> AddRelationResults
        """
        if endpoints is not None and not isinstance(endpoints, (bytes, str, list)):
            raise Exception(
                f"Expected endpoints to be a Sequence, received: {type(endpoints)}"
            )

        if via_cidrs is not None and not isinstance(via_cidrs, (bytes, str, list)):
            raise Exception(
                f"Expected via_cidrs to be a Sequence, received: {type(via_cidrs)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="AddRelation", version=20, params=_params
        )
        _params["endpoints"] = endpoints
        _params["via-cidrs"] = via_cidrs
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(AddApplicationUnitsResults)
    async def AddUnits(
        self,
        application=None,
        attach_storage=None,
        num_units=None,
        placement=None,
        policy=None,
    ):
        """AddUnits adds a given number of units to an application.

        application : str
        attach_storage : typing.Sequence[str]
        num_units : int
        placement : typing.Sequence[~Placement]
        policy : str
        Returns -> AddApplicationUnitsResults
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if attach_storage is not None and not isinstance(
            attach_storage, (bytes, str, list)
        ):
            raise Exception(
                f"Expected attach_storage to be a Sequence, received: {type(attach_storage)}"
            )

        if num_units is not None and not isinstance(num_units, int):
            raise Exception(
                f"Expected num_units to be a int, received: {type(num_units)}"
            )

        if placement is not None and not isinstance(placement, (bytes, str, list)):
            raise Exception(
                f"Expected placement to be a Sequence, received: {type(placement)}"
            )

        if policy is not None and not isinstance(policy, (bytes, str)):
            raise Exception(f"Expected policy to be a str, received: {type(policy)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="AddUnits", version=20, params=_params)
        _params["application"] = application
        _params["attach-storage"] = attach_storage
        _params["num-units"] = num_units
        _params["placement"] = placement
        _params["policy"] = policy
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationInfoResults)
    async def ApplicationsInfo(self, entities=None):
        """ApplicationsInfo returns applications information.

        entities : typing.Sequence[~Entity]
        Returns -> ApplicationInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="ApplicationsInfo", version=20, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationGetConfigResults)
    async def CharmConfig(self, args=None):
        """CharmConfig returns charm config for the input list of applications and
        model generations.

        args : typing.Sequence[~ApplicationGet]
        Returns -> ApplicationGetConfigResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="CharmConfig", version=20, params=_params
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationCharmRelationsResults)
    async def CharmRelations(self, application=None):
        """CharmRelations implements the server side of Application.CharmRelations.

        application : str
        Returns -> ApplicationCharmRelationsResults
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="CharmRelations", version=20, params=_params
        )
        _params["application"] = application
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Consume(self, args=None):
        """Consume adds remote applications to the model without creating any
        relations.

        args : typing.Sequence[~ConsumeApplicationArgV5]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="Consume", version=20, params=_params)
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def Deploy(self, applications=None):
        """Deploy fetches the charms from the charm store and deploys them
        using the specified placement directives.

        applications : typing.Sequence[~ApplicationDeploy]
        Returns -> ErrorResults
        """
        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="Deploy", version=20, params=_params)
        _params["applications"] = applications
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(DeployFromRepositoryResults)
    async def DeployFromRepository(self, args=None):
        """DeployFromRepository is a one-stop deployment method for repository
        charms. Only a charm name is required to deploy. If argument validation
        fails, a list of all errors found in validation will be returned. If a
        local resource is provided, details required for uploading the validated
        resource will be returned.

        args : typing.Sequence[~DeployFromRepositoryArg]
        Returns -> DeployFromRepositoryResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application",
            request="DeployFromRepository",
            version=20,
            params=_params,
        )
        _params["Args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(DestroyApplicationResults)
    async def DestroyApplication(self, applications=None):
        """DestroyApplication removes a given set of applications.

        applications : typing.Sequence[~DestroyApplicationParams]
        Returns -> DestroyApplicationResults
        """
        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="DestroyApplication", version=20, params=_params
        )
        _params["applications"] = applications
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def DestroyConsumedApplications(self, applications=None):
        """DestroyConsumedApplications removes a given set of consumed (remote) applications.

        applications : typing.Sequence[~DestroyConsumedApplicationParams]
        Returns -> ErrorResults
        """
        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application",
            request="DestroyConsumedApplications",
            version=20,
            params=_params,
        )
        _params["applications"] = applications
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def DestroyRelation(
        self, endpoints=None, force=None, max_wait=None, relation_id=None
    ):
        """DestroyRelation removes the relation between the
        specified endpoints or an id.

        endpoints : typing.Sequence[str]
        force : bool
        max_wait : int
        relation_id : int
        Returns -> None
        """
        if endpoints is not None and not isinstance(endpoints, (bytes, str, list)):
            raise Exception(
                f"Expected endpoints to be a Sequence, received: {type(endpoints)}"
            )

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if max_wait is not None and not isinstance(max_wait, int):
            raise Exception(
                f"Expected max_wait to be a int, received: {type(max_wait)}"
            )

        if relation_id is not None and not isinstance(relation_id, int):
            raise Exception(
                f"Expected relation_id to be a int, received: {type(relation_id)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="DestroyRelation", version=20, params=_params
        )
        _params["endpoints"] = endpoints
        _params["force"] = force
        _params["max-wait"] = max_wait
        _params["relation-id"] = relation_id
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(DestroyUnitResults)
    async def DestroyUnit(self, units=None):
        """DestroyUnit removes a given set of application units.

        units : typing.Sequence[~DestroyUnitParams]
        Returns -> DestroyUnitResults
        """
        if units is not None and not isinstance(units, (bytes, str, list)):
            raise Exception(f"Expected units to be a Sequence, received: {type(units)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="DestroyUnit", version=20, params=_params
        )
        _params["units"] = units
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def Expose(self, application=None, exposed_endpoints=None):
        """Expose changes the juju-managed firewall to expose any ports that
        were also explicitly marked by units as open.

        application : str
        exposed_endpoints : typing.Mapping[str, ~ExposedEndpoint]
        Returns -> None
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if exposed_endpoints is not None and not isinstance(exposed_endpoints, dict):
            raise Exception(
                f"Expected exposed_endpoints to be a Mapping, received: {type(exposed_endpoints)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="Expose", version=20, params=_params)
        _params["application"] = application
        _params["exposed-endpoints"] = exposed_endpoints
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationGetResults)
    async def Get(self, application=None, branch=None):
        """Get returns the charm configuration for an application.

        application : str
        branch : str
        Returns -> ApplicationGetResults
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="Get", version=20, params=_params)
        _params["application"] = application
        _params["branch"] = branch
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(CharmURLOriginResult)
    async def GetCharmURLOrigin(self, application=None, branch=None):
        """GetCharmURLOrigin returns the charm URL and charm origin the given
        application is running at present.

        application : str
        branch : str
        Returns -> CharmURLOriginResult
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if branch is not None and not isinstance(branch, (bytes, str)):
            raise Exception(f"Expected branch to be a str, received: {type(branch)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="GetCharmURLOrigin", version=20, params=_params
        )
        _params["application"] = application
        _params["branch"] = branch
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationGetConfigResults)
    async def GetConfig(self, entities=None):
        """GetConfig returns the charm config for each of the input applications.

        entities : typing.Sequence[~Entity]
        Returns -> ApplicationGetConfigResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="GetConfig", version=20, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ApplicationGetConstraintsResults)
    async def GetConstraints(self, entities=None):
        """GetConstraints returns the constraints for a given application.

        entities : typing.Sequence[~Entity]
        Returns -> ApplicationGetConstraintsResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="GetConstraints", version=20, params=_params
        )
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(StringResult)
    async def Leader(self, tag=None):
        """Leader returns the unit name of the leader for the given application.

        tag : str
        Returns -> StringResult
        """
        if tag is not None and not isinstance(tag, (bytes, str)):
            raise Exception(f"Expected tag to be a str, received: {type(tag)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="Leader", version=20, params=_params)
        _params["tag"] = tag
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def MergeBindings(self, args=None):
        """MergeBindings merges operator-defined bindings with the current bindings for
        one or more applications.

        args : typing.Sequence[~ApplicationMergeBindings]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="MergeBindings", version=20, params=_params
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def ResolveUnitErrors(self, all_=None, retry=None, tags=None):
        """ResolveUnitErrors marks errors on the specified units as resolved.

        all_ : bool
        retry : bool
        tags : Entities
        Returns -> ErrorResults
        """
        if all_ is not None and not isinstance(all_, bool):
            raise Exception(f"Expected all_ to be a bool, received: {type(all_)}")

        if retry is not None and not isinstance(retry, bool):
            raise Exception(f"Expected retry to be a bool, received: {type(retry)}")

        if tags is not None and not isinstance(tags, (dict, Entities)):
            raise Exception(f"Expected tags to be a Entities, received: {type(tags)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="ResolveUnitErrors", version=20, params=_params
        )
        _params["all"] = all_
        _params["retry"] = retry
        _params["tags"] = tags
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ScaleApplicationResults)
    async def ScaleApplications(self, applications=None):
        """ScaleApplications scales the specified application to the requested number of units.

        applications : typing.Sequence[~ScaleApplicationParams]
        Returns -> ScaleApplicationResults
        """
        if applications is not None and not isinstance(
            applications, (bytes, str, list)
        ):
            raise Exception(
                f"Expected applications to be a Sequence, received: {type(applications)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="ScaleApplications", version=20, params=_params
        )
        _params["applications"] = applications
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def SetCharm(
        self,
        application=None,
        channel=None,
        charm_origin=None,
        charm_url=None,
        config_settings=None,
        config_settings_yaml=None,
        endpoint_bindings=None,
        force=None,
        force_base=None,
        force_units=None,
        generation=None,
        resource_ids=None,
        storage_constraints=None,
    ):
        """SetCharm sets the charm for a given for the application.

        application : str
        channel : str
        charm_origin : CharmOrigin
        charm_url : str
        config_settings : typing.Mapping[str, str]
        config_settings_yaml : str
        endpoint_bindings : typing.Mapping[str, str]
        force : bool
        force_base : bool
        force_units : bool
        generation : str
        resource_ids : typing.Mapping[str, str]
        storage_constraints : typing.Mapping[str, ~StorageConstraints]
        Returns -> None
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if channel is not None and not isinstance(channel, (bytes, str)):
            raise Exception(f"Expected channel to be a str, received: {type(channel)}")

        if charm_origin is not None and not isinstance(
            charm_origin, (dict, CharmOrigin)
        ):
            raise Exception(
                f"Expected charm_origin to be a CharmOrigin, received: {type(charm_origin)}"
            )

        if charm_url is not None and not isinstance(charm_url, (bytes, str)):
            raise Exception(
                f"Expected charm_url to be a str, received: {type(charm_url)}"
            )

        if config_settings is not None and not isinstance(config_settings, dict):
            raise Exception(
                f"Expected config_settings to be a Mapping, received: {type(config_settings)}"
            )

        if config_settings_yaml is not None and not isinstance(
            config_settings_yaml, (bytes, str)
        ):
            raise Exception(
                f"Expected config_settings_yaml to be a str, received: {type(config_settings_yaml)}"
            )

        if endpoint_bindings is not None and not isinstance(endpoint_bindings, dict):
            raise Exception(
                f"Expected endpoint_bindings to be a Mapping, received: {type(endpoint_bindings)}"
            )

        if force is not None and not isinstance(force, bool):
            raise Exception(f"Expected force to be a bool, received: {type(force)}")

        if force_base is not None and not isinstance(force_base, bool):
            raise Exception(
                f"Expected force_base to be a bool, received: {type(force_base)}"
            )

        if force_units is not None and not isinstance(force_units, bool):
            raise Exception(
                f"Expected force_units to be a bool, received: {type(force_units)}"
            )

        if generation is not None and not isinstance(generation, (bytes, str)):
            raise Exception(
                f"Expected generation to be a str, received: {type(generation)}"
            )

        if resource_ids is not None and not isinstance(resource_ids, dict):
            raise Exception(
                f"Expected resource_ids to be a Mapping, received: {type(resource_ids)}"
            )

        if storage_constraints is not None and not isinstance(
            storage_constraints, dict
        ):
            raise Exception(
                f"Expected storage_constraints to be a Mapping, received: {type(storage_constraints)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="SetCharm", version=20, params=_params)
        _params["application"] = application
        _params["channel"] = channel
        _params["charm-origin"] = charm_origin
        _params["charm-url"] = charm_url
        _params["config-settings"] = config_settings
        _params["config-settings-yaml"] = config_settings_yaml
        _params["endpoint-bindings"] = endpoint_bindings
        _params["force"] = force
        _params["force-base"] = force_base
        _params["force-units"] = force_units
        _params["generation"] = generation
        _params["resource-ids"] = resource_ids
        _params["storage-constraints"] = storage_constraints
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def SetConfigs(self, args=None):
        """SetConfigs implements the server side of Application.SetConfig.  Both
        application and charm config are set. It does not unset values in
        Config map that are set to an empty string. Unset should be used for that.

        args : typing.Sequence[~ConfigSet]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="SetConfigs", version=20, params=_params)
        _params["Args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def SetConstraints(self, application=None, constraints=None):
        """SetConstraints sets the constraints for a given application.

        application : str
        constraints : Value
        Returns -> None
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if constraints is not None and not isinstance(constraints, (dict, Value)):
            raise Exception(
                f"Expected constraints to be a Value, received: {type(constraints)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application", request="SetConstraints", version=20, params=_params
        )
        _params["application"] = application
        _params["constraints"] = constraints
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def SetMetricCredentials(self, creds=None):
        """SetMetricCredentials sets credentials on the application.
        TODO (cderici) only used for metered charms in cmd MeteredDeployAPI,
        kept for client compatibility, remove in juju 4.0

        creds : typing.Sequence[~ApplicationMetricCredential]
        Returns -> ErrorResults
        """
        if creds is not None and not isinstance(creds, (bytes, str, list)):
            raise Exception(f"Expected creds to be a Sequence, received: {type(creds)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application",
            request="SetMetricCredentials",
            version=20,
            params=_params,
        )
        _params["creds"] = creds
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def SetRelationsSuspended(self, args=None):
        """SetRelationsSuspended sets the suspended status of the specified relations.

        args : typing.Sequence[~RelationSuspendedArg]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application",
            request="SetRelationsSuspended",
            version=20,
            params=_params,
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(None)
    async def Unexpose(self, application=None, exposed_endpoints=None):
        """Unexpose changes the juju-managed firewall to unexpose any ports that
        were also explicitly marked by units as open.

        application : str
        exposed_endpoints : typing.Sequence[str]
        Returns -> None
        """
        if application is not None and not isinstance(application, (bytes, str)):
            raise Exception(
                f"Expected application to be a str, received: {type(application)}"
            )

        if exposed_endpoints is not None and not isinstance(
            exposed_endpoints, (bytes, str, list)
        ):
            raise Exception(
                f"Expected exposed_endpoints to be a Sequence, received: {type(exposed_endpoints)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="Unexpose", version=20, params=_params)
        _params["application"] = application
        _params["exposed-endpoints"] = exposed_endpoints
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(UnitInfoResults)
    async def UnitsInfo(self, entities=None):
        """UnitsInfo returns unit information for the given entities (units or
        applications).

        entities : typing.Sequence[~Entity]
        Returns -> UnitInfoResults
        """
        if entities is not None and not isinstance(entities, (bytes, str, list)):
            raise Exception(
                f"Expected entities to be a Sequence, received: {type(entities)}"
            )

        # map input types to rpc msg
        _params = dict()
        msg = dict(type="Application", request="UnitsInfo", version=20, params=_params)
        _params["entities"] = entities
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def UnsetApplicationsConfig(self, args=None):
        """UnsetApplicationsConfig implements the server side of Application.UnsetApplicationsConfig.

        args : typing.Sequence[~ApplicationUnset]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application",
            request="UnsetApplicationsConfig",
            version=20,
            params=_params,
        )
        _params["Args"] = args
        reply = await self.rpc(msg)
        return reply

    @ReturnMapping(ErrorResults)
    async def UpdateApplicationBase(self, args=None):
        """UpdateApplicationBase updates the application base.
        Base for subordinates is updated too.

        args : typing.Sequence[~UpdateChannelArg]
        Returns -> ErrorResults
        """
        if args is not None and not isinstance(args, (bytes, str, list)):
            raise Exception(f"Expected args to be a Sequence, received: {type(args)}")

        # map input types to rpc msg
        _params = dict()
        msg = dict(
            type="Application",
            request="UpdateApplicationBase",
            version=20,
            params=_params,
        )
        _params["args"] = args
        reply = await self.rpc(msg)
        return reply



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to the current model
2. Deploy a local charm with a oci-image resource and waits until it reports
   itself active
3. Destroys the unit and application

"""

import asyncio
from pathlib import Path

from juju.model import Model


async def main():
    model = Model()
    print("Connecting to model")
    # connect to current model with current user, per Juju CLI
    await model.connect()

    application = None
    try:
        print("Deploying local-charm")
        base_dir = Path(__file__).absolute().parent.parent
        charm_path = f"{base_dir}/tests/integration/file-resource-charm"
        resources = {"file-res": "test.file"}
        application = await model.deploy(
            charm_path,
            resources=resources,
        )

        print("Waiting for active")
        await model.wait_for_idle()

        print("Removing Charm")
        await application.remove()
    except Exception as e:
        print(e)
        if application:
            await application.remove()
        await model.disconnect()
    finally:
        print("Disconnecting from model")
        await model.disconnect()


if __name__ == "__main__":
    asyncio.run(main())



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

# A compatibility layer on asyncio that ensures we have all the right
# bindings for the functions we need from asyncio. Reason for this
# layer is the frequent functional changes, additions and deprecations
# in asyncio across the different Python versions.

# Any module that needs to use the asyncio should get the binding from
# this layer.

import logging
import signal
import warnings
from asyncio import (
    ALL_COMPLETED as ALL_COMPLETED,
)
from asyncio import (
    FIRST_COMPLETED as FIRST_COMPLETED,
)

# FIXME: integration tests don't use these, but some are used in this repo
# Use primitives from asyncio within this repo and remove these re-exports
from asyncio import (
    Event as Event,
)
from asyncio import (
    Lock as Lock,
)
from asyncio import (
    Queue as Queue,
)
from asyncio import (
    all_tasks as all_tasks,
)
from asyncio import (
    as_completed as as_completed,
)
from asyncio import (
    create_subprocess_exec as create_subprocess_exec,
)
from asyncio import (
    current_task as current_task,
)
from asyncio import (
    ensure_future as ensure_future,
)
from asyncio import (
    gather as gather,
)
from asyncio import (
    get_event_loop_policy as get_event_loop_policy,
)
from asyncio import (
    get_running_loop as get_running_loop,
)
from asyncio import (
    new_event_loop as new_event_loop,
)
from asyncio import (
    shield as shield,
)
from asyncio import (
    sleep as sleep,
)
from asyncio import (
    subprocess as subprocess,
)
from asyncio import (
    wait,
)
from asyncio import (
    wait_for as wait_for,
)

from juju._jasyncio import (
    SingletonEventLoop as SingletonEventLoop,
)
from juju._jasyncio import (
    create_task_with_handler as create_task_with_handler,
)

warnings.warn(
    "juju.jasyncio module is being deprecated by 3.0, use asyncio or juju._jasyncio instead",
    DeprecationWarning,
    stacklevel=2,
)


ROOT_LOGGER = logging.getLogger()


def run(*steps):
    """Helper to run one or more async functions synchronously, with graceful
    handling of SIGINT / Ctrl-C.

    Returns the return value of the last function.
    """
    if not steps:
        return

    task = None
    run._sigint = False  # function attr to allow setting from closure
    # Use a singleton class to force a single event loop instance
    loop = SingletonEventLoop().loop

    def abort():
        task.cancel()
        run._sigint = True

    added = False
    try:
        loop.add_signal_handler(signal.SIGINT, abort)
        added = True
    except (ValueError, OSError, RuntimeError) as e:
        # add_signal_handler doesn't work in a thread
        if "main thread" not in str(e):
            raise
    try:
        for step in steps:
            task = loop.create_task(step)
            loop.run_until_complete(wait([task]))
            if run._sigint:
                raise KeyboardInterrupt()
            if task.exception():
                raise task.exception()
        return task.result()
    finally:
        if added:
            loop.remove_signal_handler(signal.SIGINT)



# Copyright 2023 Canonical Ltd.
# Licensed under the Apache V2, see LICENCE file for details.

"""This example:

1. Connects to current controller.
2. Loads kube config from microk8s.
3. Adds a k8s cloud.
4. Adds a userpass credential for the cloud.

"""

import asyncio
import base64
import logging
import os

import yaml

from juju.client import client
from juju.controller import Controller


async def main():
    kubecfg = os.popen("microk8s.config").read()  # noqa: S607,S605
    cfg = yaml.safe_load(kubecfg)
    ctx = {v["name"]: v["context"] for v in cfg["contexts"]}[cfg["current-context"]]
    cluster = {v["name"]: v["cluster"] for v in cfg["clusters"]}[ctx["cluster"]]
    user = {v["name"]: v["user"] for v in cfg["users"]}[ctx["user"]]

    ep = cluster["server"]
    ca_cert = base64.b64decode(cluster["certificate-authority-data"]).decode("utf-8")

    controller = Controller()
    await controller.connect()

    cloud = client.Cloud(
        auth_types=["userpass"],
        ca_certificates=[ca_cert],
        endpoint=ep,
        host_cloud_region="microk8s/localhost",
        regions=[client.CloudRegion(endpoint=ep, name="localhost")],
        type_="kubernetes",
    )
    cloud = await controller.add_cloud("test", cloud)

    cred = client.CloudCredential(
        auth_type="userpass",
        attrs={"username": user["username"], "password": user["password"]},
    )
    await controller.add_credential("test", credential=cred, cloud="test")

    await controller.remove_cloud("test")
    await controller.disconnect()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    ws_logger = logging.getLogger("websockets.protocol")
    ws_logger.setLevel(logging.INFO)
    asyncio.run(main())
